///<Add DLL>
///</Add DLL>

/************************************************************************
* Name:         Use Case: Combine Chromatic Fields Sets
* Description:  The module combines two chromatic fields sets. 
*               The user can define a lateral offset which is 
*               applied before the combination. Further a common 
*               sampling has to be defined. It is required that both 
*               sets have the same wavelengths.
*
* Authors:      Michael Kuhn (LightTrans)
* Version:      1.0 (Aug 25, 2015)
*************************************************************************/
using System;
using System.Collections.Generic;
using System.Drawing;

using LightTrans.Programming;
using VirtualLabAPI.Core.Common;
using VirtualLabAPI.Core.FieldRepresentations;
using VirtualLabAPI.Core.Numerics;
using VirtualLabAPI.Core.Modules;
using VirtualLabAPI.Core.ParameterRuns;
using VirtualLabAPI.Core.BasicFunctions;

public class VLModule : IVLModule {
    
    #region global parameters - to be defined by the user
    
    /***************************************************************/
    /*                                                             */
    /* Define U S E R   P A R A M E T E R S here                   */
    /*                                                             */
    /***************************************************************/
    
    private double _lateralOffsetOfCFS_X = 0e-3;      // lateral offset of the two CFS in x-direction
    private double _lateralOffsetOfCFS_Y = 10.0e-3;   // lateral offset of the two CFS in y-direction

    private double _samplingDistance_X = 50.0e-6;     // sampling distance of output in x-direction
    private double _samplingDistance_Y = 50.0e-6;     // sampling distance of output in y-direction

    /***************************************************************/
    /*                                                             */
    /* E N D of Definition                                         */
    /*                                                             */
    /***************************************************************/
    
    #endregion



    /// <summary>
    /// Run method of the module
    /// </summary>
    public void Run() {
        
        
        /// <summary>
        /// internal stored position vector of the chromatic fields sets
        /// </summary>
        VectorD relativePositionOfCFS = new VectorD(_lateralOffsetOfCFS_X, _lateralOffsetOfCFS_Y);
            
        /// <summary>
        /// internal stored common sampling period
        /// </summary>
        VectorD commonSamplingDistance = new VectorD(_samplingDistance_X, _samplingDistance_Y);

        CombineTwoSelectedChromaticFieldsSets(relativePositionOfCFS, commonSamplingDistance);
       
    }
    
    #region chromatic fields set combination
    /// <summary>
    /// public support method to combine two chromatic fields sets which are selected by the user
    /// </summary>
    public void CombineTwoSelectedChromaticFieldsSets(VectorD relativePositionOfCFS, VectorD commonSamplingDistance) {
        int index;
        //ask for CFS 1
        ChromaticFieldsSet cfs1 = (ChromaticFieldsSet)(Globals.ActiveDocumentHistory.BrowseLastDocuments(DocumentFilter.ChromaticFieldsSet2D,
                                                                                                         out index,
                                                                                                        "Select Chromatic Fields Set 1"));

        //ask for CFS 2
        ChromaticFieldsSet cfs2 = (ChromaticFieldsSet)(Globals.ActiveDocumentHistory.BrowseLastDocuments(DocumentFilter.ChromaticFieldsSet2D,
                                                                                                         out index,
                                                                                                         "Select Chromatic Fields Set 2"));
        //error handling
        if (cfs1 == null ||
            cfs2 == null) {
            Globals.DataDisplay.LogError("No valid CFSs selected.");
            return;
        }

        ChromaticFieldsSet cfsCombination = CalculateChromaticFieldsSetCombination(cfs1, cfs2, relativePositionOfCFS, commonSamplingDistance);
        Globals.DataDisplay.ShowDocument(cfsCombination, "Combined Chromatic Fields Set");
    }

    /// <summary>
    /// support method to convert two chromatic fields set with a lateral shift to each other
    /// </summary>
    /// <param name="cfs1">the first chromatic fields set</param>
    /// <param name="cfs2">the second chromatic fields set</param>
    /// <returns>the combined chromatic fields set</returns>
    public ChromaticFieldsSet CalculateChromaticFieldsSetCombination(ChromaticFieldsSet cfs1,
                                                                     ChromaticFieldsSet cfs2,
                                                                     VectorD relativePositionOfCFS, 
                                                                     VectorD commonSamplingDistance) {
        #region consistency check
        //check whether the chromatic fields sets are equidistant
        if (!cfs1.Stored2DDataSets.Equidistant_X ||
            !cfs1.Stored2DDataSets.Equidistant_Y ||
            !cfs2.Stored2DDataSets.Equidistant_X ||
            !cfs2.Stored2DDataSets.Equidistant_Y) {
            throw new ArgumentException("The chromatic fields sets have to be equidistant.");
        }

        //check number of subsets
        if (cfs1.Stored2DDataSets.Data.Length != cfs2.Stored2DDataSets.Data.Length) {
            throw new ArgumentException("The number of subsets within the chromatic fields do not match");
        }

        //check wavelength information
        for (int runCheckWL = 0; runCheckWL < cfs1.Wavelengths.Count; runCheckWL++) {
            if (NumericalChecks.AreInequal(cfs1.Wavelengths[runCheckWL], cfs2.Wavelengths[runCheckWL])) {
                throw new ArgumentException("The wavelengths of the chromatic fields set do not match");
            }
        }
        #endregion

        //calculate common sampling for the resulting fields
        VectorD extensionOfData = new VectorD(cfs1.Stored2DDataSets.CoordinateOfLastDataPoint_X - cfs2.Stored2DDataSets.CoordinateOfFirstDataPoint_X + relativePositionOfCFS.X,
                                              cfs1.Stored2DDataSets.CoordinateOfLastDataPoint_Y - cfs2.Stored2DDataSets.CoordinateOfFirstDataPoint_Y + relativePositionOfCFS.Y);

        SamplingParameters commonSamplingParameters = new SamplingParameters();
        commonSamplingParameters.SamplingDistance = commonSamplingDistance;
        commonSamplingParameters.SamplingPoints = (Vector)(extensionOfData / commonSamplingDistance);

        //reset interpolation method to linear due to avoid overshoots
        cfs1.Stored2DDataSets.InterpolationMethodForEquidistantSampling_X = InterpolationMethod.Linear_RealAndImaginaryPart;
        cfs1.Stored2DDataSets.InterpolationMethodForEquidistantSampling_Y = InterpolationMethod.Linear_RealAndImaginaryPart;
        cfs2.Stored2DDataSets.InterpolationMethodForEquidistantSampling_X = InterpolationMethod.Linear_RealAndImaginaryPart;
        cfs2.Stored2DDataSets.InterpolationMethodForEquidistantSampling_Y = InterpolationMethod.Linear_RealAndImaginaryPart;

        //perform data array interpolation to common sampling
        DataArray2D interpolatedDataCFS1 = cfs1.Stored2DDataSets.FieldInterpolation_PureEquidistantSampling(commonSamplingParameters);
        DataArray2D interpolatedDataCFS2 = cfs2.Stored2DDataSets.FieldInterpolation_PureEquidistantSampling(commonSamplingParameters);

        Vector relativeShiftDA1 = (Vector)(relativePositionOfCFS * new VectorD(0.5, 0.5) / commonSamplingDistance);
        Vector relativeShiftDA2 = (Vector)(relativePositionOfCFS * new VectorD(-0.5, -0.5) / commonSamplingDistance);

        //run through all members of the data array
        for (int runMemberDataArray = 0; runMemberDataArray < interpolatedDataCFS1.Data.Length; runMemberDataArray++) {       
            //apply shift on data array 1
            interpolatedDataCFS1.Data[runMemberDataArray].Shift(relativeShiftDA1);
            interpolatedDataCFS2.Data[runMemberDataArray].Shift(relativeShiftDA2);
            interpolatedDataCFS1.Data[runMemberDataArray] += interpolatedDataCFS2.Data[runMemberDataArray];
        }

        //init new chromatic fields set
        ChromaticFieldsSet cfsReturn = new ChromaticFieldsSet(interpolatedDataCFS1,
                                                              cfs1.Wavelengths,
                                                              cfs1.TypeOfChromaticFieldsSet);
                                                                         
        //set cubic interpolation for return field
        cfsReturn.Stored2DDataSets.InterpolationMethodForEquidistantSampling_X = InterpolationMethod.Cubic6P;
        cfsReturn.Stored2DDataSets.InterpolationMethodForEquidistantSampling_Y = InterpolationMethod.Cubic6P;

        return cfsReturn;
    }
    #endregion
}
